__all__ = ['Queue']

import logging
import time
import uuid

import dill
from kafka import KafkaProducer

from kq.job import Job
from kq.utils import (
    is_dict,
    is_iter,
    is_number,
    is_str,
    is_none_or_bytes,
    is_none_or_func,
    is_none_or_int,
    is_none_or_logger,
)


class Queue(object):
    """Enqueues function calls in Kafka topics as :doc:`jobs <job>`.

    :param topic: Name of the Kafka topic.
    :type topic: str
    :param producer: Kafka producer instance. For more details on producers,
        refer to kafka-python's `documentation
        <http://kafka-python.rtfd.io/en/master/#kafkaproducer>`_.
    :type producer: kafka.KafkaProducer_
    :param serializer: Callable which takes a :doc:`job <job>` namedtuple and
        returns a serialized byte string. If not set, ``dill.dumps`` is used
        by default. See :doc:`here <serializer>` for more details.
    :type serializer: callable
    :param timeout: Default job timeout threshold in seconds. If left at 0
        (default), jobs run until completion. This value can be overridden
        when enqueueing jobs.
    :type timeout: int | float
    :param logger: Logger for recording queue activities. If not set, logger
        named ``kq.queue`` is used with default settings (you need to define
        your own formatters and handlers). See :doc:`here <logging>` for more
        details.
    :type logger: logging.Logger

    **Example:**

    .. testcode::

        import requests

        from kafka import KafkaProducer
        from kq import Queue

        # Set up a Kafka producer.
        producer = KafkaProducer(bootstrap_servers='127.0.0.1:9092')

        # Set up a queue.
        queue = Queue(topic='topic', producer=producer, timeout=3600)

        # Enqueue a function call.
        job = queue.enqueue(requests.get, 'https://www.google.com/')

    .. _kafka.KafkaProducer:
        http://kafka-python.rtfd.io/en/master/apidoc/KafkaProducer.html
    """

    def __init__(self,
                 topic,
                 producer,
                 serializer=None,
                 timeout=0,
                 logger=None):

        assert is_str(topic), 'topic must be a str'
        assert isinstance(producer, KafkaProducer), 'bad producer instance'
        assert is_none_or_func(serializer), 'serializer must be a callable'
        assert is_number(timeout), 'timeout must be an int or float'
        assert timeout >= 0, 'timeout must be 0 or greater'
        assert is_none_or_logger(logger), 'bad logger instance'

        self._topic = topic
        self._hosts = producer.config['bootstrap_servers']
        self._producer = producer
        self._serializer = serializer or dill.dumps
        self._timeout = timeout
        self._logger = logger or logging.getLogger('kq.queue')
        self._default_enqueue_spec = EnqueueSpec(
            topic=self._topic,
            producer=self._producer,
            serializer=self._serializer,
            logger=self._logger,
            timeout=self._timeout,
            key=None,
            partition=None
        )

    def __repr__(self):
        """Return the string representation of the queue.

        :return: String representation of the queue.
        :rtype: str
        """
        return 'Queue(hosts={}, topic={})'.format(self._hosts, self._topic)

    def __del__(self):  # pragma: no covers
        # noinspection PyBroadException
        try:
            self._producer.close()
        except Exception:
            pass

    @property
    def hosts(self):
        """Return comma-separated Kafka hosts and ports string.

        :return: Comma-separated Kafka hosts and ports.
        :rtype: str
        """
        return self._hosts

    @property
    def topic(self):
        """Return the name of the Kafka topic.

        :return: Name of the Kafka topic.
        :rtype: str
        """
        return self._topic

    @property
    def producer(self):
        """Return the Kafka producer instance.

        :return: Kafka producer instance.
        :rtype: kafka.KafkaProducer
        """
        return self._producer

    @property
    def serializer(self):
        """Return the serializer function.

        :return: Serializer function.
        :rtype: callable
        """
        return self._serializer

    @property
    def timeout(self):
        """Return the default job timeout threshold in seconds.

        :return: Default job timeout threshold in seconds.
        :rtype: int | float
        """
        return self._timeout

    def enqueue(self, func, *args, **kwargs):
        """Enqueue a function call or a :doc:`job <job>`.

        :param func: Function or a :doc:`job <job>` object. Must be
            serializable and available to :doc:`workers <worker>`.
        :type func: callable | :doc:`kq.Job <job>`
        :param args: Positional arguments for the function. Ignored if **func**
            is a :doc:`job <job>` object.
        :param kwargs: Keyword arguments for the function. Ignored if **func**
            is a :doc:`job <job>` object.
        :return: Enqueued job.
        :rtype: :doc:`kq.Job <job>`

        **Example:**

        .. testcode::

            import requests

            from kafka import KafkaProducer
            from kq import  Job, Queue

            # Set up a Kafka producer.
            producer = KafkaProducer(bootstrap_servers='127.0.0.1:9092')

            # Set up a queue.
            queue = Queue(topic='topic', producer=producer)

            # Enqueue a function call.
            queue.enqueue(requests.get, 'https://www.google.com/')

            # Enqueue a job object.
            job = Job(func=requests.get, args=['https://www.google.com/'])
            queue.enqueue(job)

        .. note::

            The following rules apply when enqueueing a :doc:`job <job>`:

            * If ``Job.id`` is not set, a random one is generated.
            * If ``Job.timestamp`` is set, it is replaced with current time.
            * If ``Job.topic`` is set, it is replaced with current topic.
            * If ``Job.timeout`` is set, its value overrides others.
            * If ``Job.key`` is set, its value overrides others.
            * If ``Job.partition`` is set, its value overrides others.

        """
        return self._default_enqueue_spec.enqueue(func, *args, **kwargs)

    def using(self, timeout=None, key=None, partition=None):
        """Set enqueue specifications such as timeout, key and partition.

        :param timeout: Job timeout threshold in seconds. If not set, default
            timeout (specified during queue initialization) is used instead.
        :type timeout: int | float
        :param key: Kafka message key. Jobs with the same keys are sent to the
            same topic partition and executed sequentially. Applies only if the
            **partition** parameter is not set, and the producer’s partitioner
            configuration is left as default. For more details on producers,
            refer to kafka-python's documentation_.
        :type key: bytes
        :param partition: Topic partition the message is sent to. If not set,
            the producer's partitioner selects the partition. For more details
            on producers, refer to kafka-python's documentation_.
        :type partition: int
        :return: Enqueue specification object which has an ``enqueue`` method
            with the same signature as :func:`kq.queue.Queue.enqueue`.

        **Example:**

        .. testcode::

            import requests

            from kafka import KafkaProducer
            from kq import Job, Queue

            # Set up a Kafka producer.
            producer = KafkaProducer(bootstrap_servers='127.0.0.1:9092')

            # Set up a queue.
            queue = Queue(topic='topic', producer=producer)

            url = 'https://www.google.com/'

            # Enqueue a function call in partition 0 with message key 'foo'.
            queue.using(partition=0, key=b'foo').enqueue(requests.get, url)

            # Enqueue a function call with a timeout of 10 seconds.
            queue.using(timeout=10).enqueue(requests.get, url)

            # Job values are preferred over values set with "using" method.
            job = Job(func=requests.get, args=[url], timeout=5)
            queue.using(timeout=10).enqueue(job)  # timeout is still 5

        .. _documentation: http://kafka-python.rtfd.io/en/master/#kafkaproducer
        """
        return EnqueueSpec(
            topic=self._topic,
            producer=self._producer,
            serializer=self._serializer,
            logger=self._logger,
            timeout=timeout or self._timeout,
            key=key,
            partition=partition
        )


class EnqueueSpec(object):

    __slots__ = [
        '_topic',
        '_producer',
        '_serializer',
        '_logger',
        '_timeout',
        '_key',
        '_part',
        'delay'
    ]

    def __init__(self,
                 topic,
                 producer,
                 serializer,
                 logger,
                 timeout,
                 key,
                 partition):
        assert is_number(timeout), 'timeout must be an int or float'
        assert is_none_or_bytes(key), 'key must be a bytes'
        assert is_none_or_int(partition), 'partition must be an int'

        self._topic = topic
        self._producer = producer
        self._serializer = serializer
        self._logger = logger
        self._timeout = timeout
        self._key = key
        self._part = partition

    def enqueue(self, obj, *args, **kwargs):
        """Enqueue a function call or :doc:`job` instance.

        :param func: Function or :doc:`job <job>`. Must be serializable and
            importable by :doc:`worker <worker>` processes.
        :type func: callable | :doc:`kq.Job <job>`
        :param args: Positional arguments for the function. Ignored if **func**
            is a :doc:`job <job>` object.
        :param kwargs: Keyword arguments for the function. Ignored if **func**
            is a :doc:`job <job>` object.
        :return: Enqueued job.
        :rtype: :doc:`kq.Job <job>`
        """
        timestamp = int(time.time() * 1000)

        if isinstance(obj, Job):
            job_id = uuid.uuid4().hex if obj.id is None else obj.id
            func = obj.func
            args = tuple() if obj.args is None else obj.args
            kwargs = {} if obj.kwargs is None else obj.kwargs
            timeout = self._timeout if obj.timeout is None else obj.timeout
            key = self._key if obj.key is None else obj.key
            partition = self._part if obj.partition is None else obj.partition

            assert is_str(job_id), 'Job.id must be a str'
            assert callable(func), 'Job.func must be a callable'
            assert is_iter(args), 'Job.args must be a list or tuple'
            assert is_dict(kwargs), 'Job.kwargs must be a dict'
            assert is_number(timeout), 'Job.timeout must be an int or float'
            assert is_none_or_bytes(key), 'Job.key must be a bytes'
            assert is_none_or_int(partition), 'Job.partition must be an int'
        else:
            assert callable(obj), 'first argument must be a callable'
            job_id = uuid.uuid4().hex
            func = obj
            args = args
            kwargs = kwargs
            timeout = self._timeout
            key = self._key
            partition = self._part

        job = Job(
            id=job_id,
            timestamp=timestamp,
            topic=self._topic,
            func=func,
            args=args,
            kwargs=kwargs,
            timeout=timeout,
            key=key,
            partition=partition
        )
        self._logger.info('Enqueueing {} ...'.format(job))
        self._producer.send(
            self._topic,
            value=self._serializer(job),
            key=self._serializer(key) if key else None,
            partition=partition,
            timestamp_ms=timestamp
        )
        return job



__all__ = ['Message']

from collections import namedtuple

# Namedtuple which encapsulates a Kafka message.
Message = namedtuple(
    typename='Message',
    field_names=(
        'topic',      # Name of the Kafka topic (str)
        'partition',  # Topic partition (int)
        'offset',     # Offset (int)
        'key',        # Message key (bytes | None)
        'value'       # Message value (bytes)
    )
)



__all__ = ['Job']

from collections import namedtuple

# Namedtuple which encapsulates a KQ job.
Job = namedtuple(
    typename='Job',
    field_names=(
        'id',         # Job ID (str)
        'timestamp',  # Unix timestamp indicating when job was enqueued (int)
        'topic',      # Name of the Kafka topic (str)
        'func',       # Function to execute (callable)
        'args',       # Positional arguments (list)
        'kwargs',     # Keyword arguments (dict)
        'timeout',    # Job timeout threshold in seconds (int | float)
        'key',        # Kafka message key if any (str | None)
        'partition'   # Kafka topic partition if any (str | None)
    )
)

# noinspection PyUnresolvedReferences,PyProtectedMember
Job.__new__.__defaults__ = (None,) * len(Job._fields)



__all__ = ['Worker']

import _thread
import logging
import math
import threading
import traceback

import dill
from kafka import KafkaConsumer

from kq.message import Message
from kq.utils import (
    get_call_repr,
    is_str,
    is_none_or_func,
    is_none_or_logger
)


class Worker(object):
    """Fetches :doc:`jobs <job>` from Kafka topics and processes them.

    :param topic: Name of the Kafka topic.
    :type topic: str
    :param consumer: Kafka consumer instance with a group ID (required). For
        more details on consumers, refer to kafka-python's documentation_.
    :type consumer: kafka.KafkaConsumer_
    :param callback: Callback function which is executed every time a job is
        processed. See :doc:`here <callback>` for more details.
    :type callback: callable
    :param deserializer: Callable which takes a byte string and returns a
        deserialized :doc:`job <job>` namedtuple. If not set, ``dill.loads``
        is used by default. See :doc:`here <serializer>` for more details.
    :type deserializer: callable
    :param logger: Logger for recording worker activities. If not set, logger
        named ``kq.worker`` is used with default settings (you need to define
        your own formatters and handlers). See :doc:`here <logging>` for more
        details.
    :type logger: logging.Logger

    **Example:**

    .. code-block:: python

        from kafka import KafkaConsumer
        from kq import Worker

        # Set up a Kafka consumer. Group ID is required.
        consumer = KafkaConsumer(
            bootstrap_servers='127.0.0.1:9092',
            group_id='group'
        )

        # Set up a worker.
        worker = Worker(topic='topic', consumer=consumer)

        # Start the worker to process jobs.
        worker.start()

    .. _documentation:
        http://kafka-python.rtfd.io/en/master/#kafkaconsumer
    .. _kafka.KafkaConsumer:
        http://kafka-python.rtfd.io/en/master/apidoc/KafkaConsumer.html
    """

    def __init__(self,
                 topic,
                 consumer,
                 callback=None,
                 deserializer=None,
                 logger=None):

        assert is_str(topic), 'topic must be a str'
        assert isinstance(consumer, KafkaConsumer), 'bad consumer instance'
        assert consumer.config['group_id'], 'consumer must have group_id'
        assert is_none_or_func(callback), 'callback must be a callable'
        assert is_none_or_func(deserializer), 'deserializer must be a callable'
        assert is_none_or_logger(logger), 'bad logger instance'

        self._topic = topic
        self._hosts = consumer.config['bootstrap_servers']
        self._group = consumer.config['group_id']
        self._consumer = consumer
        self._callback = callback
        self._deserializer = deserializer or dill.loads
        self._logger = logger or logging.getLogger('kq.worker')

    def __repr__(self):
        """Return the string representation of the worker.

        :return: String representation of the worker.
        :rtype: str
        """
        return 'Worker(hosts={}, topic={}, group={})'.format(
            self._hosts, self._topic, self._group
        )

    def __del__(self):  # pragma: no cover
        # noinspection PyBroadException
        try:
            self._consumer.close()
        except Exception:
            pass

    def _execute_callback(self, status, message, job, res, err, stacktrace):
        """Execute the callback.

        :param status: Job status. Possible values are "invalid" (job could not
            be deserialized or was malformed), "failure" (job raised an error),
            "timeout" (job timed out), or "success" (job finished successfully
            and returned a result).
        :type status: str
        :param message: Kafka message.
        :type message: :doc:`kq.Message <message>`
        :param job: Job object, or None if **status** was "invalid".
        :type job: kq.Job
        :param res: Job result, or None if an exception was raised.
        :type res: object | None
        :param err: Exception raised by job, or None if there was none.
        :type err: Exception | None
        :param stacktrace: Exception traceback, or None if there was none.
        :type stacktrace: str | None
        """
        if self._callback is not None:
            try:
                self._logger.info('Executing callback ...')
                self._callback(status, message, job, res, err, stacktrace)
            except Exception as e:
                self._logger.exception(
                    'Callback raised an exception: {}'.format(e))

    def _process_message(self, msg):
        """De-serialize the message and execute the job.

        :param msg: Kafka message.
        :type msg: :doc:`kq.Message <message>`
        """
        self._logger.info(
            'Processing Message(topic={}, partition={}, offset={}) ...'
            .format(msg.topic, msg.partition, msg.offset))
        try:
            job = self._deserializer(msg.value)
            job_repr = get_call_repr(job.func, *job.args, **job.kwargs)

        except Exception as err:
            self._logger.exception('Job was invalid: {}'.format(err))
            self._execute_callback('invalid', msg, None, None, None, None)
        else:
            self._logger.info('Executing job {}: {}'.format(job.id, job_repr))

            if job.timeout:
                timer = threading.Timer(job.timeout, _thread.interrupt_main)
                timer.start()
            else:
                timer = None
            try:
                res = job.func(*job.args, **job.kwargs)
            except KeyboardInterrupt:
                self._logger.error(
                    'Job {} timed out or was interrupted'.format(job.id))
                self._execute_callback('timeout', msg, job, None, None, None)
            except Exception as err:
                self._logger.exception(
                    'Job {} raised an exception:'.format(job.id))
                tb = traceback.format_exc()
                self._execute_callback('failure', msg, job, None, err, tb)
            else:
                self._logger.info('Job {} returned: {}'.format(job.id, res))
                self._execute_callback('success', msg, job, res, None, None)
            finally:
                if timer is not None:
                    timer.cancel()

    @property
    def hosts(self):
        """Return comma-separated Kafka hosts and ports string.

        :return: Comma-separated Kafka hosts and ports.
        :rtype: str
        """
        return self._hosts

    @property
    def topic(self):
        """Return the name of the Kafka topic.

        :return: Name of the Kafka topic.
        :rtype: str
        """
        return self._topic

    @property
    def group(self):
        """Return the Kafka consumer group ID.

        :return: Kafka consumer group ID.
        :rtype: str
        """
        return self._group

    @property
    def consumer(self):
        """Return the Kafka consumer instance.

        :return: Kafka consumer instance.
        :rtype: kafka.KafkaConsumer
        """
        return self._consumer

    @property
    def deserializer(self):
        """Return the deserializer function.

        :return: Deserializer function.
        :rtype: callable
        """
        return self._deserializer

    @property
    def callback(self):
        """Return the callback function.

        :return: Callback function, or None if not set.
        :rtype: callable | None
        """
        return self._callback

    def start(self, max_messages=math.inf, commit_offsets=True):
        """Start processing Kafka messages and executing jobs.

        :param max_messages: Maximum number of Kafka messages to process before
            stopping. If not set, worker runs until interrupted.
        :type max_messages: int
        :param commit_offsets: If set to True, consumer offsets are committed
            every time a message is processed (default: True).
        :type commit_offsets: bool
        :return: Total number of messages processed.
        :rtype: int
        """
        self._logger.info('Starting {} ...'.format(self))

        self._consumer.unsubscribe()
        self._consumer.subscribe([self.topic])

        messages_processed = 0
        while messages_processed < max_messages:
            record = next(self._consumer)

            message = Message(
                topic=record.topic,
                partition=record.partition,
                offset=record.offset,
                key=record.key,
                value=record.value
            )
            self._process_message(message)

            if commit_offsets:
                self._consumer.commit()

            messages_processed += 1

        return messages_processed



import logging
from inspect import ismethod, isfunction, isbuiltin, isclass


def get_call_repr(func, *args, **kwargs):
    """Return the string representation of the function call.

    :param func: A callable (e.g. function, method).
    :type func: callable
    :param args: Positional arguments for the callable.
    :param kwargs: Keyword arguments for the callable.
    :return: String representation of the function call.
    :rtype: str
    """
    # Functions, builtins and methods
    if ismethod(func) or isfunction(func) or isbuiltin(func):
        func_repr = '{}.{}'.format(func.__module__, func.__qualname__)
    # A callable class instance
    elif not isclass(func) and hasattr(func, '__call__'):
        func_repr = '{}.{}'.format(func.__module__, func.__class__.__name__)
    else:
        func_repr = repr(func)

    args_reprs = [repr(arg) for arg in args]
    kwargs_reprs = [k + '=' + repr(v) for k, v in sorted(kwargs.items())]
    return '{}({})'.format(func_repr, ', '.join(args_reprs + kwargs_reprs))


def is_none_or_logger(obj):
    return obj is None or isinstance(obj, logging.Logger)


def is_none_or_int(obj):
    return obj is None or isinstance(obj, int)


def is_none_or_bytes(obj):
    return obj is None or isinstance(obj, bytes)


def is_none_or_func(obj):
    return obj is None or callable(obj)


def is_str(obj):
    return isinstance(obj, str)


def is_number(obj):
    return isinstance(obj, (int, float))


def is_dict(obj):
    return isinstance(obj, dict)


def is_iter(obj):
    return isinstance(obj, (list, tuple))



__all__ = ['Job', 'Message', 'Queue', 'Worker', '__version__']

from kq.job import Job
from kq.message import Message
from kq.queue import Queue
from kq.worker import Worker
from kq.version import __version__
